; (function () {

    // 工具类
    const util = {
        /**
         * 
         * @returns 生成唯一标识
         */
        getUUID: function () {
            return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
                var r = (Math.random() * 16) | 0,
                    v = c == 'x' ? r : (r & 0x3) | 0x8;
                return v.toString(16);
            });
        },

        /**
         * 计算签名
         * @param {*} salt  盐值
         * @param {*} randomKey 随机数
         */
        sign: function (salt, randomKey) {
            let saltArray = Array.from(salt);
            let randomArray = Array.from(randomKey);

            let min = Math.min(saltArray.length, randomArray.length);
            let sum = 0;
            for (let i = 0; i < min; i++) {
                let codeSalt = saltArray[0].charCodeAt(0);
                let codeRandom = randomArray[0].charCodeAt(0);

                sum += codeSalt;
                sum += codeRandom;
            }

            let result = sum % 10;
            return result;
        },

        /**
         * 获取选择器
         * @param {*} path  路径
         * @returns  结果
         * @since 0.0.2
         */
        getSelectors: function (path) {
            // 反转 + 过滤 + 映射 + 拼接
            return path
                .reverse()
                .filter((element) => {
                    return element !== document && element !== window;
                })
                .map((element) => {
                    //console.log("element", element.nodeName);
                    let selector = "";
                    if (element.id) {
                        return `${element.nodeName.toLowerCase()}#${element.id}`;
                    } else if (element.className && typeof element.className === "string") {
                        return `${element.nodeName.toLowerCase()}.${element.className}`;
                    } else {
                        selector = element.nodeName.toLowerCase();
                    }
                    return selector;
                })
                .join(" ");
        },

        /**
         * 获取选择器
         * @param {*} pathsOrTarget  路径或者对象
         * @since 0.0.2
         */
        getSelector: function (pathsOrTarget) {
            if (Array.isArray(pathsOrTarget)) {
                return util.getSelectors(pathsOrTarget);
            } else {
                let path = [];
                while (pathsOrTarget) {
                    path.push(pathsOrTarget);
                    pathsOrTarget = pathsOrTarget.parentNode;
                }
                return util.getSelectors(path);
            }
        },

        /**
         * 获取选择的行信息
         * @param {*} stack  堆栈
         * @returns 结果
         * @since 0.0.2
         */
        getLines: function (stack) {
            return stack
                .split("\n")
                .slice(1)
                .map((item) => item.replace(/^\s+at\s+/g, ""))
                .join("^");
        },

        /**
         * 时间格式化
         * @param {*} time  时间
         * @returns 结果
         * @since 0.0.3
         */
        formatTime: function (time) {
            // 格式化
            return time;
        },

        /**
         * 获取选择器的名称
         * 
         * @param {*} element 元素
         * @since 0.0.5
         */
        getSelectElementName: function(element) {
            if (element.id) {
                return "#" + element.id;
            } else if (element.className) {// a b c => .a.b.c
                return "." + element.className.split(' ').filter(item => !!item).join('.');
            } else {
                return element.nodeName ? element.nodeName.toLowerCase() : "";
            }
        },

        /**
         * 是否为包裹的元素
         * @param {*} element 元素
         * @since 0.0.5
         */
        isWrapperElem: function(element) {
            //包裹的元素
            let wrapperElements = ['html', 'body', '#container', '.content'];
            let selector = util.getSelectElementName(element);
            if (wrapperElements.indexOf(selector) != -1) {
                return true;
            }

            return false;
        },

        /**
         * 中划线改驼峰
         * @param {*} name 中划线名称
         */
        middleToCamel: function (name) {
            let arr = [...str];
            arr.forEach((item, index) => {
                if (item === "-") {
                    arr.splice(index, 1);
                    arr[index] = arr[index].toUpperCase();
                }
            });
            return arr.join('');
        },

        /**
         * 鼠标事件标准化
         * 
         * https://www.jianshu.com/p/e2d8231ff9fe
         * @param {*} event 事件
         * @since 0.7.0
         */
        mouseEventStandard: function(event) {
            if(!event){  // 兼容IE浏览器
                event = window.event;
                event.target = event.srcElement;
                event.layerX = event.offsetX;
                event.layerY = event.offsetY;
            }

            // 计算鼠标指针的x轴距离
            event.mouseX = event.pageX || event.clientX + document.body.scrollLeft;
            // 计算鼠标指针的y轴距离
            event.mouseY = event.pageY || event.clientY + document.body.scrollTop;
            // 返回标准化的事件对象
            return event;  
        },
    };

    // 定义跟踪器
    const track = {
        // 基础配置
        clientConfig: {
            // 基本配置
            serverUrl: "http://localhost:8080/monitor-web-server/track",
            version: "1.0.0",
            appId: 'test',
            appSecret: '0000000000111111111111',
            // 页面维度的 sessionId，每次页面加载初始化
            pageSessionId: '',
            // 全局维度的 sessionId，只初始化一次
            globalSessionId : '',

            // 当前操作的用户信息
            userId : '',
            userName : '',

            // 功能开关
            jsPromiseErrorFlag: true,
            jsResourceErrorFlag: true,
            pvFlag: true,
            stayTimeFlag: true,
            injectXHRFlag: true,
            blankScreenFlag: true,

            // 性能相关
            elementPerfFlag: true,
            firstInputPerfFlag: true,
            largestContentfulPaintPerfFlag: true,
            layoutShiftPerfFlag: true,
            longtaskPerfFlag: true,
            navigationPerfFlag: true,
            paintPerfFlag: true,
            resourcePerfFlag: true,
            eventPerfFlag: false,
            markPerfFlag: false,
            measurePerfFlag: false,

            // ----------------------------------- 鼠标事件
            // 单击鼠标左键时发生，如果右键也按下则不会发生。当用户的焦点在按钮上并按了 Enter 键时，同样会触发这个事件
            clickFlag: true,
            //双击鼠标左键时发生，如果右键也按下则不会发生
            dblclickFlag: true,
            //单击任意一个鼠标按钮时发生 
            mousedownFlag: false,
            //松开任意一个鼠标按钮时发生
            mouseupFlag: false,
            //鼠标指针位于某个元素上且将要移出元素的边界时发生
            mouseoutFlag: false,
            //鼠标指针移出某个元素到另一个元素上时发生
            mouseoverFlag: false,
            //鼠标在某个元素上时持续发生
            mousemove: false,    

            // ------------------------------------ 键盘事件
            //在键盘上按下某个键时触发。
            keydownFlag: false,
            keyupFlag: false,
            keypressFlag: true,
        },

        /**
         * 开始时间
         */
        pvStartTime: 0,

        // 上一次的事件
        lastEvent: {},

        // 页面生命周期
        pageLife: {
            DOMContentLoaded: function () {
                // 初始化唯一的 sessionId，用于标识当前一个页面的标识
                track._initPageSessionId();
            },
            // 页面加载
            load: function () {

                // 更新 pv 开始时间
                this.pvStartTime = Date.now();

                // 启用 pv 统计
                track.pv();

                // 加载完成之后，判断是否为白屏
                if (document.readyState === 'complete') {
                    track.blankScreen();
                } 
            },
            beforeunload: function (event) {
                // console.log('Page beforeunload')

                // const beforeunloadOptions = {}
                // track.send(beforeunloadOptions)
            },
            unload: function () {
                // 在这里可以使用navigator.sendBeacon(url, data)发送异步传输
                // console.log('Page unloaded')

                // 统计用户停留时间
                track.stayTime();
            }
        },

        /**
         * 白色的屏幕
         * @since 0.0.5
         */
        blankScreen: function() {
            if(!this.clientConfig.blankScreenFlag) {
                return;
            }
            
            let emptyPoints = 0;

            // 计算总数，共计 18 个点
            for (let i = 1; i <= 9; i++) {
                let xElements = document.elementsFromPoint(window.innerWidth * i / 10, window.innerHeight / 2);
                let yElements = document.elementsFromPoint(window.innerWidth / 2, window.innerHeight * i / 10);

                if(util.isWrapperElem(xElements[0])) {
                    emptyPoints++;
                }
                if(util.isWrapperElem(yElements[0])) {
                    emptyPoints++;
                }
            }

            // 此处认为超过 16 个就是白屏，是否可以配置?
            // ps: 感觉这种统计方法不是很准确
            if(emptyPoints >= 16) {
                let centerElements = document.elementsFromPoint(window.innerWidth / 2, window.innerHeight / 2);
                let elemName = util.getSelectElementName(centerElements);

                let blankScreenData = {
                    kind: 'stability',
                    type: 'blankScreen',
                    emptyPoints: emptyPoints,
                    selector: elemName
                };

                // 发送消息
                track.send(blankScreenData);
            }    
        },

        /**
         * 开始 pv 统计
         * @since 0.0.3
         */
        pv: function () {
            if (!this.clientConfig.pvFlag) {
                return;
            }

            let connection = navigator.connection;
            const loadOptions = {
                kind: "business",   // 业务大类
                type: "pv",     //业务子类
                effectiveType: connection.effectiveType, //网络环境
                rtt: connection.rtt, //往返时间
            };

            track.send(loadOptions);
        },

        /**
         * 用户停留时间
         * @since 0.0.3
         */
        stayTime: function () {
            if (!this.clientConfig.stayTimeFlag) {
                return;
            }

            let stayTime = Date.now() - this.pvStartTime;
            const unloadOptions = {
                kind: "business",
                type: "stayTime",   // 停留时间
                stayTime: stayTime
            };
            track.send(unloadOptions);
        },

        /**
         * 后端请求报错监控
         * @since 0.0.4
         */
        injectXHR: function () {
            if(!this.clientConfig.injectXHRFlag) {
                return;
            }

            let serverUrl = this.clientConfig.serverUrl;
            let XMLHttpRequest = window.XMLHttpRequest;
            let oldOpen = XMLHttpRequest.prototype.open;
            XMLHttpRequest.prototype.open = function (method, url, async) {
                // 把上报接口过滤掉（避免死循环）
                if (url === serverUrl) {
                    this.logData = null;
                } else {
                    this.logData = { method, url, async };
                }
                
                return oldOpen.apply(this, arguments);
            };

            let oldSend = XMLHttpRequest.prototype.send;
            XMLHttpRequest.prototype.send = function (body) {
                if (this.logData != null) {
                    let startTime = Date.now();
                    // 事件状态变化
                    this.addEventListener("readystatechange", function (event) {
                        switch (xhr.readyState) {
                            case 0:
                                // console.log(0, '未初始化...');
                                break;
                            case 1:
                                // console.log(1, '请求参数已准备，尚未发送请求...');
                                break;
                            case 3:
                                // console.log(3, '已接收数据长度:' + xhr.responseText.length);
                                break;
                            case 4:
                                // console.log(4, '响应全部接受完毕');
                                if ((xhr.status >= 200 && xhr.status < 300) || xhr.status == 304) {
                                    // success
                                } else {
                                    // fail 的场景
                                    let duration = Date.now() - startTime;
                                    let status = xhr.status;
                                    let statusText = xhr.statusText;

                                    const xhrData = {
                                        kind: "stability",
                                        type: "xhr",
                                        eventType: 'error',
                                        pathname: this.logData.url,
                                        status: status + "-" + statusText, // 状态码
                                        duration: duration,
                                        response: this.response ? JSON.stringify(this.response) : "", // 响应体
                                        params: body || "", // 入参
                                    };

                                    // 发送消息
                                    track.send(xhrData);
                                }
                                break;
                        }
                    }, false);
                }
                return oldSend.apply(this, arguments);
            };
        },

        /**
         * 文件加载异常获取 js 语法错误等
         * @since 0.0.3
         */
        jsResourceError: function () {
            if (!this.clientConfig.jsResourceErrorFlag) {
                return;
            }

            // 异常处理
            window.addEventListener("error", (event) => {
                console.log("error+++++++++++", event);

                let lastEvent = this.lastEvent; // 获取到最后一个交互事件
                // 脚本加载错误
                if (event.target && (event.target.src || event.target.href)) {
                    const jsErrorData = {
                        kind: "stability", // 监控指标的大类，稳定性
                        type: "error", // 小类型，这是一个错误
                        errorType: "resourceError", // js执行错误
                        filename: event.target.src || event.target.href, // 哪个文件报错了
                        tagName: event.target.tagName,
                        selector: util.getSelector(event.target), // 代表最后一个操作的元素    
                    };
                    track.send(jsErrorData);
                } else {
                    const jsErrorData = {
                        kind: "stability", // 监控指标的大类，稳定性
                        type: "error", // 小类型，这是一个错误
                        errorType: "jsError", // js执行错误
                        message: event.message, // 报错信息
                        filename: event.filename, // 哪个文件报错了
                        position: `${event.lineno}:${event.colno}`, // 报错的行列位置
                        stack: event.error ? util.getLines(event.error.stack) : "",
                        selector: lastEvent ? util.getSelector(lastEvent.path) : "", // 代表最后一个操作的元素
                    };
                    track.send(jsErrorData);
                }
            }, true);
        },

        /**
         * JS 执行异常
         * @since 0.0.3
         */
        jsPromiseError: function () {
            if (!this.clientConfig.jsPromiseErrorFlag) {
                return;
            }

            window.addEventListener(
                "unhandledrejection",
                (event) => {
                    console.log("unhandledrejection-------- ", event);
                    // 这里其实不太准确，应该事件一直在变化，有没有更加准确的方法呢？
                    // 获取最近的多个操作？
                    let lastEvent = this.lastEvent; // 获取到最后一个交互事件
                    let message;
                    let filename;
                    let line = 0;
                    let column = 0;
                    let stack = "";
                    let reason = event.reason;
                    if (typeof reason === "string") {
                        message = reason;
                    } else if (typeof reason === "object") {
                        message = reason.message;
                        if (reason.stack) {
                            let matchResult = reason.stack.match(/at\s+(.+):(\d+):(\d+)/);
                            filename = matchResult[1];
                            line = matchResult[2];
                            column = matchResult[3];
                        }
                        stack = util.getLines(reason.stack);
                    }

                    const rejectData = {
                        kind: "stability", // 监控指标的大类，稳定性
                        type: "error", // 小类型，这是一个错误
                        errorType: "promiseError", // js执行错误
                        message:message, // 报错信息
                        filename:filename, // 哪个文件报错了
                        position: `${line}:${column}`, // 报错的行列位置
                        stack:stack,
                        selector: lastEvent ? util.getSelector(lastEvent.path) : "", // 代表最后一个操作的元素
                    };
                    track.send(rejectData);
                },
                true
            );
        },

        /**
         * 性能监控
         * @since 0.0.5
         */
        performance: function() {
            let allArrays = PerformanceObserver.supportedEntryTypes;

            for(let i = 0; i < allArrays.length; i++) {
                let eventName = allArrays[i];

                let propName = eventName + 'PerfFlag';
                if(!this.clientConfig[propName]) {
                    continue;
                }

                // 处理
                let observerArrays = [eventName];

                // 性能测试
                new PerformanceObserver(function(list, observer) {
                    list.getEntries().forEach((entry) => {
                        
                        const perfData = {
                            kind: "performance",
                            entryType: entry.entryType,
                            name: entry.name,
                            startTime: util.formatTime(entry.startTime), // 开始时间
                            duration: util.formatTime(entry.duration), // 持续时间
                        };

                        track.send(perfData);
                    });

                    observer.disconnect(); // 不再观察了
                }).observe({entryTypes: observerArrays});    
            }
        },

        /**
         * 初始化鼠标的事件
         * @since 0.0.6
         */
        initMouseEvent() {
            let events = ["click", "dbclick", "mousedown", "mouseout", "mouseover", "mouseup", "mousemove"];

            //遍历事件
            events.forEach((e)=>{
                let propName = e+'Flag';

                if(!this.clientConfig[propName]) {
                    return;
                }

                // 添加对应的监听器
                document.addEventListener(e,
                    function (event) {
                        let eventStandard = util.mouseEventStandard(event);

                        let mouseEventData = {
                            kind: "business",
                            mouseX: eventStandard.mouseX,
                            mouseY: eventStandard.mouseY,
                            type: eventStandard.type,
                            pointerType: eventStandard.pointerType,
                            selector: util.getSelector(eventStandard.target)
                        };

                        track.send(mouseEventData);
                    },
                    {
                        capture: true, // 是在捕获阶段还是冒泡阶段执行
                        passive: true, // 默认不阻止默认事件
                    }
                );  
            })
        },

        /**
         * 初始化键盘事件
         * @since 0.8.0
         */
        initKeyEvent() {
            let events = ["keydown", "keyup", "keypress"];

            //遍历事件
            events.forEach((e)=>{
                let propName = e+'Flag';

                if(!this.clientConfig[propName]) {
                    return;
                }

                // 添加对应的监听器
                document.addEventListener(e,
                    function (event) {
                        // 兼容 IE
                        let target = event.target ? event.target : event.srcElement;

                        let eventData = {
                            kind: "business",
                            type: event.type,
                            key: event.key,
                            keyCode: event.keyCode,
                            charCode: event.charCode,
                            shiftKey: event.shiftKey,
                            metaKey: event.metaKey,
                            altKey: event.altKey,

                            selector: util.getSelector(target)
                        };

                        track.send(eventData);
                    },
                    {
                        capture: true, // 是在捕获阶段还是冒泡阶段执行
                        passive: true, // 默认不阻止默认事件
                    }
                );  
            })
        },

        /**
         * 数据上报
         * @param { String } data
         */
        send: function (data) {
            let actualData = this._parseParam(data);

            // 计算唯一标识+签名
            let body = JSON.stringify(actualData);

            // TODO: xhr 发送
            // console.log("config", JSON.stringify(this.clientConfig));
            console.log("body", body);

            this.xhr = new XMLHttpRequest();
            this.xhr.open("POST", this.clientConfig.serverUrl, true);
            this.xhr.setRequestHeader("Content-Type", "application/json");
            // this.xhr.setRequestHeader("x-log-apiversion", "1.0.0");
            // this.xhr.setRequestHeader("x-log-bodyrawsize", body.length);
            this.xhr.onload = function () {
                // console.log(this.xhr.response);
            };
            this.xhr.onerror = function (error) {
                // console.log(error);
            };
            this.xhr.send(body);
        },

        /**
         * 参数编码返回字符串
         */
        _parseParam: function (param) {
            // 公共参数拼接
            return Object.assign({}, this._getExtraData(), param);
        },

        /**
         * 
         * @returns 获取额外参数
         */
        _getExtraData: function () {
            let uuid = util.getUUID();

            // 经纬度一般有限制，不建议使用
            // navigator.geolocation.getCurrentPosition(res=>{
            //     console.log(`------------------------res`, res)
            //     // res=> GeolocationPosition{
            //     //		coords:{	
            //     //			accuracy: 20185 // 准确度
            //     //			altitude: null  // 高度
            //     //			altitudeAccuracy: null
            //     //			heading: null
            //     //			latitude: 22.3193039 // 纬度
            //     //			longitude: 114.16936109999999 // 经度
            //     //		},
            //     //		timestamp: 1619592517698, // 时间戳
            //     // }
            // });

            return {
                domain: document.domain || '',  //获取域名    
                title: document.title  || '',   //页面标题
                url: location.href || '',       //页面地址
                referrer: document.referrer || '',  //上一跳路径

                screenHeight: window.screen.height || 0,    //显示屏高度
                screenWidth: window.screen.width || 0,  //显示屏宽度
                screenColorDepth: window.screen.colorDepth || 0,
                viewHeight: window.innerHeight || 0, //内视图高度
                viewWidth: window.innerWidth || 0, //内视图宽度

                timestamp: Date.now(),  // 当前时间戳
                userAgent: navigator.userAgent || '', // 浏览器内核
                lang: navigator.language || '', // 语言
                cookie: document.cookie,  //cookie 信息

                uuid: uuid, // 请求的唯一标识

                appId: this.clientConfig.appId,
                version: this.clientConfig.version,
                ex: util.sign(this.clientConfig.appSecret, uuid),

                pageSessionId: track._getPageSessionId(),
                globalSessionId: this.clientConfig.globalSessionId,
                userId: this.clientConfig.userId,
                userName: this.clientConfig.userName,
            };
        },

        /**
         * 设置对应的 sessionId 标识
         */
        _setSessionId(sessionId) {
            // 设置唯一的标识
            sessionId = sessionId || util.getUUID();
            console.log("设置对应的 sessionId: " + sessionId);
            this.clientConfig.sessionId = sessionId;
        },

        /**
         * 初始化声明周期
         * TODO: 监听的时间，添加对应的开关配置
         */
        _initLifeCycle() {
            Object.keys(this.pageLife).forEach((hook) => {
                // console.log("初始化声明周期: " + hook);
                window.addEventListener(hook, this.pageLife[hook])
            })
        },

        /**
         * 初始化全局 sessionId，也可以放在内存中。
         * 
         * @since 0.9.0
         */
        _initGlobalSessionId() {
            let key = '__MONITOR_WEB_JS_GSID';
            let gsid = localStorage.getItem(key) || util.getUUID();

            this.clientConfig.globalSessionId = gsid;
            localStorage.setItem(key, gsid);
        },

        /**
         * 设置对应的用户信息
         * @param {*} userId 用户标识
         * @param {*} userName 用户名称
         */
        _setUserInfo: function (userId, userName) {
            this.clientConfig.userId = userId;
            this.clientConfig.userName = userName;
        },

        /**
         * 初始化页面的 sessionId
         * @since 0.9.0
         */
        _initPageSessionId() {
            this.clientConfig.pageSessionId = util.getUUID();
        },

        /**
         * 获取对应的 sessionId
         * @since 0.9.0
         */
        _getPageSessionId() {
            return this.clientConfig.pageSessionId || util.getUUID();
        },
    };

    // 暴露公共方法
    window.MonitorWeb = {
        /**
         * 自定义事件上报
         * @param {*} params 上报数据
         */
        init: function (params) {
            // 配置初始化
            params = params || {};
            track.clientConfig = Object.assign({}, track.clientConfig, params);

            // 初始化全局的标识
            track._initGlobalSessionId();

            // 绑定声明周期初始化
            track._initLifeCycle();

            // js 异常处理
            track.jsResourceError();
            track.jsPromiseError();

            // XHR 请求
            track.injectXHR();

            // 性能相关
            track.performance();

            // 鼠标
            track.initMouseEvent();

            // 键盘
            track.initKeyEvent();
        },

        /**
         * 设置对应的用户信息
         * @param {*} userId 用户标识
         * @param {*} userName 用户名称
         */
        setUserInfo: function (userId, userName) {
            track._setUserInfo(userId, userName);
        }
    };

})();